#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from opendeco import tracker, mix
from opendeco.units import DepthUnit

class DiveSeries(object):
   
    dives = []
   
    def addDive(self, dive):
        assert isinstance(dive, DiveProfile)
        self.dives.append(dive)

    def __iter__(self):
        for dive in self.dives:
            yield dive
            
class DiveProfile(object):
    
    surfaceInterval = 0
    segments = None

    def __init__(self):
        self.segments = []

    def __repr__(self):
        s = "<DiveProfile>\n"
        for seg in self.segments:
            s += "  %s\n" % repr(seg)
        s += "</DiveProfile>"
        return s
   
    def totalTime(self):
        t = 0.0
        for s in self.segments:
            t += s.time
        return t

    def addSegment(self, seg):
        assert isinstance(seg, SegmentBase)
        if self.segments:
            # assert that there are no discontinuities in the profile
            assert self.segments[-1].endDepth.bar == seg.startDepth.bar
        else:
            # first segment has to start at a depth of zero
            assert seg.startDepth.bar==0
        self.segments.append(seg)
        t = self.totalTime()

    def __iter__(self):
        for seg in self.segments:
            yield seg


class SegmentBase(object):
    """
    Base class for defining a segment of a dive profile. A segment
    can be a part where the diver levels or a part where
    the diver changes depth.
    """
    mix = None
    startDepth = None
    endDept = None
    time = None
    
    def __repr__(self):
        s = "<%s mix=%s start=%s end=%s time=%s>" % (self.__class__.__name__,
            self.mix, self.startDepth, self.endDepth, self.time)
        return s
    
class SegmentChange(SegmentBase):
    
    rate = None

    def __init__(self, gasmix, start, end, rate):
        assert isinstance(start, DepthUnit)
        assert isinstance(end, DepthUnit)
        assert isinstance(rate, DepthUnit)
        assert isinstance(gasmix, mix.GasMix)
        self.mix = gasmix
        self.startDepth = start
        self.endDepth = end
        self.rate = rate
        self.time = abs( 1.0 * (end.bar - start.bar) / rate.bar )
   
class SegmentLevel(SegmentBase):
   
    depth = None

    def __init__(self, gasmix, depth, dt):
        assert isinstance(depth, DepthUnit)
        assert isinstance(gasmix, mix.GasMix)
        self.mix = gasmix
        self.depth = depth
        self.startDepth = depth
        self.endDepth = depth
        self.time = dt
    

if __name__=='__main__':
    import pylab
    from numpy import *
    p = DiveProfile()
    p.addSegment(SegmentChange(None, 0, 30, 15))
    p.addSegment(SegmentLevel(None, 30, 40))
    p.addSegment(SegmentChange(None, 30, 20, 10))
    p.addSegment(SegmentLevel(None, 20, 10))
    p.addSegment(SegmentChange(None, 20, 0, 5))
    d = tracker.current.plotdata('profile.depth')
    pylab.plot(d.x, -d.y, 'o-')
    #pylab.savefig('tmp/test.png')
    #pylab.savefig('tmp/test.eps')
    pylab.show()


