#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
FIXME
- uitleg waarom de aanname dat luchtdruk 1 bar is gebruikelijk en veilig is
- rename to gasmix.py
"""

from opendeco.units import DepthUnit, msw, bar

class GasMix(object):
    
    # oxygen part
    o2 = 0.21
    # helium part
    he = 0.00
    # nitrogen part
    n2 = 0.79

    o2_pct = 21
    he_pct = 0
    n2_pct = 79

    def __init__(self, o2=0.21, he=0.00):
        assert o2 <= 1.00
        assert o2 > 0.00
        assert he < 1.00
        assert he >= 0.00
        self.o2 = o2
        self.he = he
        self.n2 = 1.00 - o2 - he
        assert self.n2 < 1.00
        assert self.n2 >= 0.00
        self.o2_pct = int(round(self.o2 * 100))
        self.he_pct = int(round(self.he * 100))
        self.n2_pct = int(round(self.n2 * 100))

    def name(self):
        o2 = self.o2_pct
        he = self.he_pct
        n2 = self.n2_pct
        if o2==100:
            return "Oxygen"
        elif o2==21 and n2==79:
            return "Air"
        elif he==0:
            return "Nitrox %i" % o2
        elif n2==0:
            return "Heliox %i" % o2
        else:
            return "Trimix %i/%i" % (o2, he)

    def shortname(self):
        o2 = self.o2_pct
        he = self.he_pct
        n2 = self.n2_pct
        if o2==100:
            return "O2"
        elif o2==21 and n2==79:
            return "Air"
        elif he==0:
            return "EAN%i" % o2
        else:
            return "Tx%i/%i" % (o2, he)

    def __str__(self):
        return self.name()

    def __repr__(self):
        return "<GasMix O2=%.2f He=%.2f>" % (self.o2_pct, self.he_pct)

    def EAD(self, depth):
        """
        Calculates the Equivalent Air Depth 
        Does not work for trimix!
        depth   - DepthUnit instance
        """
        assert isinstance(depth, DepthUnit)
        if self.he:
            raise "EAD() does not work for trimix!"
        return ( ( (1.0-self.o2) * (depth.bar + 1.0) ) / 0.79 ) - 1.0

    def PO2(self, depth):
        """
        Calculates the Partial O2 pressure
        depth   - DepthUnit instance
        """
        assert isinstance(depth, DepthUnit)
        return self.o2 * (depth.bar + 1.0)

    def PO2depth(self, po2):
        """
        Calculates the depth with given Partial O2 pressure
        po2 - partial O2 pressure (bar)
        returns DepthUnit (bar)
        """
        return (( po2 / self.o2 ) - 1.0) * bar

    def FIXME_TOD(self):
        """
        FIXME
        Calculates the maximum safe depth the gas can be breathed at
        (1.4 bar partial O2 pressure)
        """
        return (14/self.o2) - 10

    def FIXME_MOD(self):
        """
        FIXME
        Calculates the maximum safe depth the gas can be breathed at
        (1.6 bar partial O2 pressure)
        """
        return (16/self.o2) - 10

Air = GasMix(0.21)
Oxygen = GasMix(1.00)

def EAN(o2_pct):
    return GasMix(o2_pct/100.0)

def Tx(o2_pct, he_pct):
    return GasMix(o2_pct/100.0, he_pct/100.0)


def idealMix(depth, po2=1.4, END=30*msw):
    """
    Calculates the ideal mixture for a given depth.
    Assumption: surface pressure is 1 bar, will add safety otherwise.
    po2 max Partial O2 pressure
    END max Equivalent Nitrogen Depth
    FIXME: how calculated
    """
    assert isinstance(depth, DepthUnit)
    assert isinstance(END, DepthUnit)
    o2 = po2 / (depth.bar + 1.0)
    o2 = min(1.0, o2)
    n2 = 0.79 * (END.bar + 1.0) / (depth.bar + 1.0)
    n2 = min(1.0-o2, n2)
    he = 1 - o2 - n2
    return GasMix(o2, he)


if __name__=="__main__":
    import sys
    if len(sys.argv)>1:
        f = float(sys.argv[1])
        if f<1.0:
            mix = GasMix(f)
            d = int(sys.argv[2])
            print "EAD of %im is %.1f" % (d, mix.EAD(d))
            print "PO2 of %im is %.2f" % (d, mix.PO2(d))
        else:
            mix = idealGasMix(f)
            print "Ideal mixture for %im: o2=%.2f he=%.2f n2=%.2f" % \
                (f,mix.o2,mix.he,mix.n2)
            mix = idealGasMix(f, END=60*msw)
            print "Ideal nitrox for %im: o2=%.2f he=%.2f n2=%.2f" % \
                (f,mix.o2,mix.he,mix.n2)
            print "EAD of %.0fm is %.1f" % (f, mix.EAD(f))
            #print "Bottom Marge is %.1f" % mix.bottomMarge()
    else:
        air = GasMix()
        n32 = GasMix(0.32)
        n36 = GasMix(0.36)
        print "EAD of 30m using EAN32 is %.1f" % n32.EAD(30*msw)
        print "PO2 of 30m using EAN32 is %.2f" % n32.PO2(30*msw)
        print "PO2 is 1.0 using EAN32 at depth %.1f" % n32.PO2depth(1.0).msw
        #print "Bottom Max using air is %.1f" % air.bottomMax()
        #print "Bottom Max using EAN32 is %.1f" % n32.bottomMax()
        #print "Bottom Max using EAN36 is %.1f" % n36.bottomMax()
        #print "Bottom Marge using air is %.1f" % air.bottomMarge()
        #print "Bottom Marge using EAN32 is %.1f" % n32.bottomMarge()
        for depth in (7,18,38,76):
            mix = idealMix(depth*msw)
            print "Ideal mixture for %im: o2=%.2f he=%.2f n2=%.2f" % \
                (depth, mix.o2, mix.he, mix.n2)
        gasses = [(.21,), (.32,), (.11,.50), (.20, .80), (.18, .45)]
        for g in gasses:
            if len(g)==1:
                m = GasMix(g[0])
            else:
                m = GasMix(g[0], g[1])
            print g, m.name(), m.shortname()

