#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import logging
logger = logging.getLogger('opendeco.tissues')

from numpy import array, exp, log

from opendeco.diveseries import SegmentLevel, SegmentChange
from opendeco.tracker import current
from opendeco.vpmphysics import const

halftimes = {}

# the 15 VPM tissue halftimes (minutes) in the YH paper
# the Yount and Hoffman paper says the 1min tissues is never used
# WARNING: no helium halftimes are defined, using same values!
yh_n2 = array([  1,   2,   5,  10,  20,  40,  80, 120, 
               160, 240, 320, 400, 480, 560, 720])
halftimes['yh'] = (yh_n2, yh_n2)

# the 16 Buhlmann halftimes, also used in Erik Baker's Fortran implementation VPMDECO
buhlmann_he = array([ 1.88,  3.02,  4.72,  6.99,  10.21,  14.48,  20.53,  29.11,
                     41.20, 55.19, 70.69, 90.34, 115.29, 147.42, 188.24, 240.03])
buhlmann_n2 = array([  5.0,   8.0,  12.5,  18.5,   27.0,   38.3,   54.3,   77.0,
                     109.0, 146.0, 187.0, 239.0,  305.0,  390.0,  498.0,  635.0])
halftimes['buhlmann'] = (buhlmann_he, buhlmann_n2)


def tissueLoadingConstantDepth(P_t0, P_alv, dt, t_h, trackid=None):
    """
    Calculates new tissue loading after a stay a constant depth.
    This is known as the Haldane equation.
    http://www.deepocean.net/deepocean/index.php?science03.php
    P_alv   : (float) inert gas partial pressure in lungs
    P_t0    : (array/float) tissue loading at start of segment
    t_h     : (array/float) halftimes for inert gas
    returns : (array/float) new tissue loadings
    """
    k = log(2) / t_h
    
    def f(t):
        return P_alv + ( P_t0 - P_alv ) * exp(-k * t)
    
    P_t = f(dt)
    
    # for debugging and plotting purposes
    if trackid and current.resolution:
        t = 0.0
        while t<dt:
            # FIXME: doesn't work properly if we call it in a loop
            current.log(trackid, t, f(t))
            t += current.resolution

    return f(dt)


def tissueLoadingLinearlyChangingDepth(P_t0, P_alv0, P_alv1, dt, t_h, runtime=0, 
                                       trackid=None):
    """
    Calculates new tissue loading after a lineair change of depth
    This is also known as the Schreiner equation
    P_alv0  : (float) inert gas partial pressure in lungs at begin
    P_alv1  : (float) inert gas partial pressure in lungs at end
    P_t0    : (array/float) tissue loading at start of segment
    t_h     : (array/float) halftimes for inert gas
    runtime : (float) current runtime before segment (for logging)
    returns : (array/float) new tissue loadings
    """
    if not dt>0:
        return P_t0
    
    k = log(2) / t_h
    rate = 1.0 * (P_alv1 - P_alv0) / dt
    
    def f(t):
        return P_alv0 + rate*(t-(1/k)) - ( P_alv0 - P_t0 - rate/k ) * exp(-k * t)
    
    # for debugging and plotting purposes
    if trackid and current.resolution:
        t = 0.0
        while t<dt:
            # FIXME: doesn't work properly if we call it in a loop
            current.log(trackid, t+runtime, f(t))
            t += current.resolution

    return f(dt)


def startOfDecoZone(mix, depth, ascentRate, tissuesHe, tissuesN2, surfacePressure, ht_he, ht_n2):
    """
    Calculates the depth at which during ascent the gradient becomes positive
    and thus the compartment enters the decozone. From this point on bubbles can 
    be excited into growth.
    returns [array] depth of start of deco zone for each compartment (bar)
    """
    # inpired partial pressures
    ambient = depth.bar + surfacePressure
    insp_he = (ambient - const.waterVaporPressure.bar) * mix.he
    insp_n2 = (ambient - const.waterVaporPressure.bar) * mix.n2
    # rate of change in partial pressures
    rate_he = ascentRate.bar * mix.he
    rate_n2 = ascentRate.bar * mix.n2
    # bounds for time (from current depth to vacuum of space)
    low = 0.0
    high = ambient / ascentRate.bar

    # FIXME: check these
    # these can be determined all at once
    low_gradient = tissuesHe + tissuesN2 + const.pressureOtherGasses.bar - ambient
    insp_end_he = insp_he - high * rate_he
    insp_end_n2 = insp_n2 - high * rate_n2
    high_he = tissueLoadingLinearlyChangingDepth(tissuesHe, insp_he, insp_end_he, high, ht_he)
    high_n2 = tissueLoadingLinearlyChangingDepth(tissuesN2, insp_n2, insp_end_n2, high, ht_n2)
    high_gradient = high_he + high_n2 + const.pressureOtherGasses.bar
    
    size = len(ht_he)
    deco_start = array([0.0] * size)
    for i in range(size):
        high_g = high_gradient[i]
        low_g = low_gradient[i]

        # do some checks
        if high_g * low_g >= 0.0:
            print high_g, low_g
            raise Exception("Root is not within brackets (already in deco zone??)")

        if low_g < 0:
            t = low
            diff_change = high - low
        else:
            t = high
            diff_change = low - high

        iteration = 0
        while 1:
            last_diff_change = diff_change
            diff_change = 0.5 * last_diff_change

            mid = t + diff_change

            insp_mid_he = insp_he - mid * rate_he
            insp_mid_n2 = insp_n2 - mid * rate_n2
            mid_he_i = tissueLoadingLinearlyChangingDepth(tissuesHe[i], insp_he, \
                insp_mid_he, mid, ht_he[i])
            mid_n2_i = tissueLoadingLinearlyChangingDepth(tissuesN2[i], insp_n2, \
                insp_mid_n2, mid, ht_n2[i])
            mid_gradient_i = mid_he_i + mid_n2_i + const.pressureOtherGasses.bar \
                - ambient + ascentRate.bar * mid

            if mid_gradient_i <= 0.0:
                t = mid

            if abs(diff_change) < 1e-3 or mid_gradient_i==0.0:
                # converged
                break

            iteration += 1
            if iteration > 150:
                raise Exception("Exceeded max of 150 iterations")

        deco_start[i] = ambient - ascentRate.bar * t - surfacePressure

    return deco_start



if __name__=='__main__':
    from pylab import *
    ht = yh_n2
    tis = array([0.79] * len(ht))
    # descend in 2 min FIXME: fout
    tis = tissueLoadingLinearlyChangingDepth(tis, 0.79*1.0, 0.79*4.0, 2.0, ht)
    print tis
    current.timedelta += 2
    # dive at 30m for 20min
    tis = tissueLoadingConstantDepth(tis, 0.79*4.0, 2.0, ht)
    d = current.plotdata('tissue.load')
    #print d.x
    #print d.y[:,0]
    for i in range(len(ht)):
        plot(d.x, -d.y[:,i])
    show()
    

