#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Tracking of datapoints
"""

from numpy import array, max, min

current = None

class PlotData(object):
    x = None
    y = None
    comment = None
    annotation = None
    ndatasets = 0

    def __init__(self, data=[]):
        x = []
        y = []
        self.comment = []
        self.annotation = []
        for d in data:
            try:
                l = list(d[1])
                self.ndatasets = min(len(l), self.ndatasets)
            except TypeError:
                self.ndatasets = 0
            x.append(d[0])
            y.append(d[1])
            self.comment.append(d[2])
            self.annotation.append(d[3])
        self.x = array(x)
        self.y = array(y)

    def dataset(self, i):
        """
        Get a certain subset of Y-datapoints, for example a specific 
        tissues compartment if the datapoint are compartment arrays.
        """
        if self.ndatasets==0:
            raise TypeError('Datapoints in Y are not arrays')
        elif i >= self.ndatasets:
            raise IndexError('Y has datapoints 0-%i, you tried to access %i' \
                % (self.ndatasets-1, i))
        else:
             return self.y[:,i]

    @property
    def xrange(self):
        return min(self.x), max(self.x)

    @property
    def yrange(self):
        return min(self.y), max(self.y)

    def get(self, x):
        """
        Get y for given x (doing linear interpolation)
        Assumes x[n] < x[n+1]
        FIXME: test
        """
        x0 = None
        x1 = None
        prev = None
        idx = 0
        # find the two surrounding points
        for cur in self.x:
            if x <= cur:
                # we've just past our lower point
                x0 = prev
                x1 = cur
                break
            prev = cur
            idx += 1
        if idx==0:
            raise ValueError("Requested x is lower than lowest x, will not extrapolate")
        if x0 is None:
            raise ValueError("Requested x is higher than highest x, will not extrapolate")
        y0 = self.y[idx-1]
        y1 = self.y[idx]
        dydx = 1.0 * (y1 - y0) / (x1 - x0)
        y = y0 + (x - x0) * dydx
        return y


class Tracker(object):
    
    # Hint about time resolution, may be ignored by the code that
    # logs the data. If None no intermediate points will be calculated.
    # resolution = 0.2
    resolution = None

    # add this to reported time
    timedelta = 0.0

    # dump every log point to screen
    debug = False

    # dict of id:list
    # list contains tuples: (time, value, comment, annotation)
    data = None

    def __init__(self):
        self.data = {}

    def log(self, id, time, value, comment=None, annotation=None):
        """Logs a datapoint.
        id         : string identifier of parameter, example: 'tissue3.po2'
        time       : time in minutes of event
        value      : value to log, should be numerical
        comment    : optional comment, could be used for string log messages
        annotation : anotation for datapoint for use in plots
        """
        if not self.data.has_key(id):
            self.data[id] = []
        datapoints = self.data[id]
        datapoints.append( (time + self.timedelta, value, comment, annotation) )
        if self.debug:
            print 'Tracker %s:' % id, datapoints[-1]

    def datapoints(self, id):
        for p in self.data[id]:
            yield (p[0], p[1])

    def plotdata(self, id):
        return PlotData(self.data[id])

current = Tracker()

