#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

#<vpm
import math

import logging
logger = logging.getLogger('opendeco.vpm')

from numpy import array, exp, maximum

from opendeco import tissues, diveseries, mix
from opendeco.units import mmHg, msw, bar, atm
from opendeco.tracker import current
from opendeco.vpmphysics import const, radiusAfterGradient, crushingPressure, \
    initialAllowableGradient, newAllowableGradient, ceilings, surfacePhaseVolumeTime, \
    boylesLawCompensation, radiusRegeneration, conservatismRadiusFactor, newAllowableGradient2

class DecoAlgorithm(object):
    
    def __setattr__(self, key, val):
        # only allow to set already defined attributes to prevent dangerous typos
        # raises an AttributeError if the attribute doesn't exist
        getattr(self, key)
        object.__setattr__(self, key, val)
    

class VPM(DecoAlgorithm):

    # tissues halftimes (ht_he, ht_n2)
    halftimes = tissues.halftimes['buhlmann']

    # number of different tissue compartments
    tissueCount = 0
    
    # array of tissue loadings (bar)
    # FIXME: need seperate for He
    # FIXME: tensions noemen?
    tissuesN2 = None
    tissuesHe = None

    # holds the maximum crushing pressure for each compartment
    # is only updated on descends
    maxPCrushHe = None
    maxPCrushN2 = None
    
    # array of critical radii
    radiusN2 = None
    radiusHe = None

    # array filled with 1.0 of right dimension
    one = None

    # time in dive (min)
    runtime = 0

    # used for deco calculation (DepthUnit per min)
    ascentSpeed = 10 * msw

    # gas breathed at surface
    #surfaceMix = mix.Air
    surfaceMix = mix.Air

    # pressure at surface
    surfacePressure = 1.013 * bar
    #surfacePressure = 1.0

    # minutes to simulate after the dive ended (min)
    simulateAfterDive = 30

    # deco stop depth interval
    decoDepthInterval = 3 * msw

    # align stop times to this (min) (1.0 means that ascends to the next stop will occur 
    # every time the runtime rollsover to the next minute)
    alignStopTimes = 1.0

    # depth of last deco stop (many people prefer 4.5 or 6 msw
    # when diving with rough surface conditions)
    # FIXME: implement
    decoLastDepth = 3 * msw

    # decompression gas mixes (excluding bottom gas)
    # list of (switchdepth as DepthUnit, mix)
    # decoMixes = [(6*msw, mix.Oxygen)]
    decoMixes = None

    # last gas mix used before deco (bottom mix for first stage of decompression)
    lastGasMix = surfaceMix

    # current depth (used as a starting point for deco calculation)
    currentDepth = 0 * msw

    # current depth (used as a starting point for deco calculation)
    lastMaxDepth = 0 * msw

    # number of critical volume iterations for last decompression profile
    lastIterationCount = 0

    # depth at which the deco zone started in previous dive
    lastStartDecoZone = 0 * msw

    # first deco stop depth in previous dive
    lastFirstStop = 0 * msw

    # deco schedule of last dive
    # [ textlines ]
    lastScheduleText = None
    # [ (depth, stoptime, runtime_to_leave, mix) ]
    lastSchedule = None

    # use Boyle's Law compensation? (VPM-B)
    useBoylesLawCompensation = True

    # use regeneration of radii (has an effect on very long dives)
    useRegeneration = True

    # use the iterative Critical Volume Algorithm
    # (disabling it will basicly set the allowed excess volume to zero
    #  and thus yield very conservative schedules)
    useCriticalVolumeAlgorithm = True

    # V-Planner style conservatism setting (int 0-5)
    # This will enlarge the initial critical radii by a certain amount.
    # WARNING: Must be set through constructor (or 'reset()' should 
    #          be called after changing it). Cannot be changed between dives!
    conservatism = 0

    # number of dives executed
    numberOfDives = 0

    # which algorithm?
    # 1 - Baker 2 - new one
    useAllowableGradient = 1

    def __init__(self, **kwargs):
        ht = self.halftimes
        assert len(ht[0]) == len(ht[1]), "Number of He and N2 compartments differs"
        self.tissueCount = len(ht[0])
        self.one = array( [1.0] * self.tissueCount )
        self.decoMixes = []
        for key, val in kwargs.items():
            getattr(self, key)
            setattr(self, key, val)
        self.reset()


    def reset(self): #:reset
        #<stateBefore
        # initialize tissues with steady state surface condition
        p_insp = self.surfacePressure.bar - const.waterVaporPressure.bar
        self.tissuesN2 = array( [p_insp * self.surfaceMix.n2] \
            * self.tissueCount )
        self.tissuesHe = array( [p_insp * self.surfaceMix.he] \
            * self.tissueCount )
        
        #<conservatism
        # initialize critical radii
        crf = conservatismRadiusFactor(self.conservatism)
        self.radiusN2 = array( [const.initialCriticalRadiusN2 * crf] * self.tissueCount )
        self.radiusHe = array( [const.initialCriticalRadiusHe * crf] * self.tissueCount )
        #>conservatism

        # initialize max crushing pressures
        self.maxPCrushHe = array( [0.0] * self.tissueCount )
        self.maxPCrushN2 = array( [0.0] * self.tissueCount )
        #>stateBefore

    #<segmentTissue
    def segment(self, mix, p0, p1, time, simulate=False): #:segment
        """
        Execute a dive segment:
        - adjust tissue loading
        - update maximum bubble crushing pressure on descent
        - adjust runtime to reflect this segment
        mix : [GasMix] gas breathed for this segment
        p0 : (bar) absolute ambient pressure at start 
        p1 : (bar) absolute ambient pressure at end
        time : (min) time duration of segment 
        simulate : whether to update this instance to reflect the changes
            if True, this call will have no effect on further calculation
        Returns: nothing
        """
        # helium
        P_alv0 = (p0 - const.waterVaporPressure.bar) * mix.he
        P_alv1 = (p1 - const.waterVaporPressure.bar) * mix.he
        ht = self.halftimes[0]
        tis_he = tissues.tissueLoadingLinearlyChangingDepth(
            self.tissuesHe, P_alv0, P_alv1, time, ht, self.runtime,
            trackid="tissue.load.he")

        # nitrogen
        P_alv0 = (p0 - const.waterVaporPressure.bar) * mix.n2
        P_alv1 = (p1 - const.waterVaporPressure.bar) * mix.n2
        ht = self.halftimes[1]
        tis_n2 = tissues.tissueLoadingLinearlyChangingDepth(
            self.tissuesN2, P_alv0, P_alv1, time, ht, self.runtime,
            trackid="tissue.load.n2")
        #>segmentTissue
        
        #<segmentCrush
        if p1 > p0:
            # descent, calculate crushing pressure and update maximums
            # for this we need:
            # - initial gas tensions (array)
            ten0 = self.tissuesN2 + self.tissuesHe + const.pressureOtherGasses.bar
            current.log('tissue.tension', self.runtime, ten0)
            # - ending gas tensions (array)
            ten1 = tis_n2 + tis_he + const.pressureOtherGasses.bar
            current.log('tissue.tension', self.runtime+time, ten1)
            cp_he, cp_n2 = crushingPressure(p0, p1, ten0, ten1)
            current.log('crushingpressure.last.he', self.runtime+time, cp_he)
            current.log('crushingpressure.last.n2', self.runtime+time, cp_n2)
            # set maximum values
            logger.debug('max crush before %s' % self.maxPCrushN2)
            self.maxPCrushHe = maximum(self.maxPCrushHe, cp_he)
            self.maxPCrushN2 = maximum(self.maxPCrushN2, cp_n2)
            current.log('crushingpressure.max.he', self.runtime+time, self.maxPCrushHe)
            current.log('crushingpressure.max.n2', self.runtime+time, self.maxPCrushN2)
            logger.debug('max crush after %s' % self.maxPCrushN2)
        #>segmentCrush

        if not simulate:
            # to make it possible to calculate datapoints after the dive
            # without actually affecting the current state
            self.tissuesN2 = tis_n2
            self.tissuesHe = tis_he
            self.runtime += time
            self.lastGasMix = mix
            self.currentDepth = (p1 - self.surfacePressure.bar) * bar
            self.lastMaxDepth = max(self.lastMaxDepth, self.currentDepth)
            current.log('profile.depth', self.runtime, p1-self.surfacePressure.bar)

    def runProfile(self, dive): #:runProfile
        """Process dive"""
        logger.info("Running profile")
        self.numberOfDives += 1
        self.currentDepth = 0 * msw
        self.lastMaxDepth = 0 * msw
        current.log('profile.depth', 0, 0)
        
        # simulate a diver working on his tan on the surface
        p = self.surfacePressure.bar
        self.segment(self.surfaceMix, p, p, dive.surfaceInterval)

        # reset runtime for dive
        self.runtime = 0.0
        for seg in dive:
            if isinstance(seg, diveseries.SegmentLevel) or \
                     isinstance(seg, diveseries.SegmentChange):
                p0 = self.surfacePressure.bar + seg.startDepth.bar
                p1 = self.surfacePressure.bar + seg.endDepth.bar
                if seg.startDepth < seg.endDepth:
                    logstr = "[%.1f min] Descent from %s to %s takes %s min" % \
                        (self.runtime, seg.startDepth, seg.endDepth, seg.time)
                elif seg.startDepth > seg.endDepth:
                    logstr = "[%.1f min] Ascent from %s to %s takes %s min" % \
                        (self.runtime, seg.startDepth, seg.endDepth, seg.time)
                else:
                    logstr = "[%.1f min] Level at %s for %s min" % \
                        (self.runtime, seg.startDepth, seg.time)
                logger.info(logstr)
                self.segment(seg.mix, p0, p1, seg.time) #:callSegment
            else:
                raise Exception("unknown segment type")

        # decompression
        self.decompress()
        
        # calculate more datapoints for some time after the dive to be able
        # to create nice tissue loading plots
        self.segment(self.surfaceMix, p, p, self.simulateAfterDive, simulate=True)


    #<decompress
    #<nocva
    def decompress(self):
        """Calculate deco schedule for an ascend
        returns: DiveProfile
        
        Instead of a complicated approach with the use of modifications of
        the Schreiner equation to predict various events like the first deco 
        stop, we will use a simpler more robust method. We'll just take the 
        decompression step by step and see wat happens along the way. This
        makes it a lot easier to implement gas switches, special preferences
        for deco stop extensions, etc.
        
        At the end of the last diving segment, we will ascend to the first 
        deco depth step, typically a multiple of 3 meters. So when our last 
        depth was 40m, we will ascend to 39m. From there on we will extend
        our stop while a ceiling prevents us from ascending to the next one.
        Along the way we will do gas switches and all other stuff that has 
        to be done, like extending certain stops for whatever purpose.
        """
        # radius regeneration
        if self.useRegeneration:
            self.maxPCrushHe = radiusRegeneration(self.maxPCrushHe, self.radiusHe, 
                self.runtime)
            self.maxPCrushN2 = radiusRegeneration(self.maxPCrushN2, self.radiusN2, 
                self.runtime)
        
        # calculate new radii with max crushing pressures
        # FIXME: mag ik deze updaten?
        logger.debug("Radii before gradient: %s %s" % (self.radiusHe, self.radiusN2))
        self.radiusN2 = radiusAfterGradient(self.radiusN2, self.maxPCrushN2)
        self.radiusHe = radiusAfterGradient(self.radiusHe, self.maxPCrushHe)
        logger.debug("Radii after gradient: %s %s" % (self.radiusHe, self.radiusN2))
        current.log('radius.crushed.he', self.runtime, self.radiusHe)
        current.log('radius.crushed.n2', self.runtime, self.radiusN2)
        
        # iteration counter (from 1, not 0)
        iteration = 1

        # copies for deco iterations, wrapped in class because Python doesn't allow
        # setting variables from nested functions
        class temp:
            tissuesHe = self.tissuesHe
            tissuesN2 = self.tissuesN2
            runtime = self.runtime
            maxActualGradients = array([0.0] * len(self.tissuesHe))
        temp = temp()

        # Calculate initial allowable gradients
        # These will be updated in several iterations
        # When Critical Volume is disabled, this will also be the final ones.
        initialAllowHe = initialAllowableGradient(self.radiusHe)
        initialAllowN2 = initialAllowableGradient(self.radiusN2)
        allowHeFirstStop = initialAllowHe
        allowHe = initialAllowHe
        allowN2FirstStop = initialAllowN2
        allowN2 = initialAllowN2
        logger.debug("Initial allowable gradients: %s %s" % (allowHe, allowN2))
        current.log('gradient.allowed.he', iteration, allowHe)
        current.log('gradient.allowed.n2', iteration, allowN2)
        #>nocva
        
        def currentCeiling():
            """
            Calculate current ceiling
            """
            # uses: allowHe, allowN2
            ceils = ceilings(temp.tissuesHe, temp.tissuesN2, 
                            allowHe, allowN2, const.pressureOtherGasses.bar)
            logger.debug("Compartment ceilings: %s" % (ceils))
            controlling = ceils.argmax()
            ceil_bar = ceils[controlling]
            ceil_depth = (ceil_bar - self.surfacePressure.bar) * bar
            logger.info("Controlling compartment (from 0) #%i, ceiling %s bar (%.2fm)" % (
                controlling, ceil_bar, ceil_depth.msw))
            return ceil_depth

        #<nocva
        # calculate start of decozone, the depth at which the gradient in one or more
        # compartments can excite bubbles of critical radius into growth
        deco_start_all = tissues.startOfDecoZone(self.lastGasMix, self.currentDepth, \
            self.ascentSpeed, self.tissuesHe, self.tissuesN2, self.surfacePressure.bar, \
            self.halftimes[0], self.halftimes[1])
        deco_start = max(deco_start_all) * bar
        self.lastStartDecoZone = deco_start
        time_to_deco_start = (self.currentDepth.bar - deco_start.bar) / self.ascentSpeed.bar
        logger.info("Start of deco zone: %.2fm" % deco_start.msw)
        
        # calculate first possible stop depth using the decoDepthInterval
        firststop_n = int( math.floor( self.currentDepth.bar / self.decoDepthInterval.bar ) )
        firststop = firststop_n * self.decoDepthInterval
        ceil_depth = currentCeiling()
        if firststop < ceil_depth:
            # this shouldn't happen unless decoDepthInterval is set too high
            raise Exception("First possible deco stop shallower than ceiling, "
                "decoDepthInterval too large?")
        sp = self.surfacePressure.bar

        # Now let's ascend to the first possible stop depth so we
        # can enter our deco loop.
        # TODO: if a gasswith is planned from t=0 (unlikely) it happens
        #       after this segment instead of before
        dt = (self.currentDepth.bar - firststop.bar) / self.ascentSpeed.bar
        self.segment(self.lastGasMix, self.currentDepth.bar+sp, firststop.bar+sp, dt)
        #>nocva

        def decosegment(mix, p0, p1, time): #:decosegment
            """
            Segment calculation without updating global values
            Uses and updates the following outer scope variables:
            - temp.tissuesHe
            - temp.tissuesN2
            - temp.runtime
            - temp.maxActualGradients
            Uses read only:
            - iteration
            - self
            """
            # calculate inspired pressures
            P_alv0_he = (p0 - const.waterVaporPressure.bar) * mix.he
            P_alv1_he = (p1 - const.waterVaporPressure.bar) * mix.he
            P_alv0_n2 = (p0 - const.waterVaporPressure.bar) * mix.n2
            P_alv1_n2 = (p1 - const.waterVaporPressure.bar) * mix.n2
            # calulate gradient at start
            g0 = const.pressureOtherGasses.bar + temp.tissuesHe + temp.tissuesN2 - p0
            
            # helium
            ht = self.halftimes[0]
            temp.tissuesHe = tissues.tissueLoadingLinearlyChangingDepth(
                temp.tissuesHe, P_alv0_he, P_alv1_he, time, ht, temp.runtime,
                trackid="tissue.deco.iter%i.he" % iteration)
            # nitrogen
            ht = self.halftimes[1]
            temp.tissuesN2 = tissues.tissueLoadingLinearlyChangingDepth(
                temp.tissuesN2, P_alv0_n2, P_alv1_n2, time, ht, temp.runtime,
                trackid="tissue.deco.iter%i.n2" % iteration)

            # calulate gradient at end
            g1 = const.pressureOtherGasses.bar + temp.tissuesHe + temp.tissuesN2 - p1
            # update maximum gradient
            temp.maxActualGradients = maximum(temp.maxActualGradients, g0)
            temp.maxActualGradients = maximum(temp.maxActualGradients, g1)
            current.log('tissue.deco.iter%i.gradient' % iteration, 
                temp.runtime, g0)
            current.log('tissue.deco.iter%i.gradient' % iteration, 
                temp.runtime + time, g1)
            logger.debug("Gradients at start, runtime=%.2f: %s" % (temp.runtime, g0))
            logger.debug("Gradients at end, runtime=%.2f: %s" % (temp.runtime + time, g1))
            # log profile
            current.log('profile.deco.iter%i.depth' % iteration, 
                temp.runtime, p1 - self.surfacePressure.bar)
            temp.runtime += time
            return

        #<nocva
        # prepare for Critical Volume iteration
        converged = False
        decotime = 1e99
        # get list of decomixes and sort by depth
        decomixes = sorted(self.decoMixes, key=lambda x: x[0].msw)
        # for predictable comparison of depth floats
        epsilon = 0.05
        while not converged:
            # this is a nice place to make a copy of variables before iteration
            temp.tissuesHe = self.tissuesHe[:]
            temp.tissuesN2 = self.tissuesN2[:]
            temp.runtime = self.runtime
            temp.maxActualGradients = array([0.0] * len(self.tissuesHe))
            mix = self.lastGasMix
            mixes = decomixes[:]
            schedule = []
            scheduleText = []
            firststop = None
            allowHe= allowHeFirstStop
            allowN2= allowN2FirstStop

            # Now that that's done let's start our decompression loop.
            # This basicly keeps adding time until we can ascend to the next stop.
            # Everytime we arrive at a new depth we check if we need to switch gasses,
            # if the time at this depth should be extended for some reason, etc.
            for stop_n in range(firststop_n, 0, -1):
                thisstop = stop_n * self.decoDepthInterval
                nextstop = (stop_n-1) * self.decoDepthInterval
                
                # check for gas switch
                while mixes and (mixes[-1][0].msw + epsilon) > thisstop.msw:
                    # switch to next one
                    mix = mixes.pop()[1]
                    po2 = mix.PO2(thisstop)
                    logger.info("Gas switch to %s at %.1f m (PO2=%.2f)" % (mix, 
                        thisstop.msw, po2))
                
                ceil_depth = currentCeiling() #:callCurrentCeiling
                logger.info("[%.1f min] Now at %.2fm, ceiling %.2fm" % \
                    (temp.runtime, thisstop.msw, ceil_depth.msw))

                if ceil_depth.bar > thisstop.bar:
                    # this shouldn't ever happen, but it can't hurt to check
                    raise Exception, \
                        "We've broken through our ceiling during ascent to next stop!"
                
                if ceil_depth.bar <= nextstop.bar:
                    # don't stop here, keep ascending
                    pass
                else:
                    if firststop is None:
                        firststop = thisstop
                        if firststop.bar > deco_start:
                            raise Exception("First stop deeper that start of decozone")
                    elif self.useBoylesLawCompensation:
                        # Boyle's Law compensation of allowed gradients (VPM-B)
                        allowHe = boylesLawCompensation(firststop, nextstop, \
                            self.surfacePressure.bar, allowHeFirstStop)
                        allowN2 = boylesLawCompensation(firststop, nextstop, \
                            self.surfacePressure.bar, allowN2FirstStop)
                    #current.log('tissue.deco.iter%i.allowedgradient' % iteration, \
                    #    temp.runtime, allow

                    stoptime = 0.0
                    # TODO take into account offgassing in ascent (now slightly more conservative)
                    while ceil_depth.bar > nextstop.bar:
                        # stay at this depth until we can ascend to the next stop
                        dt = self.alignStopTimes - (temp.runtime % self.alignStopTimes)
                        logger.debug("[%5.1f min] depth %.2fm, ceiling %.2fm, hang %.2fmin" % \
                            (temp.runtime, thisstop.msw, ceil_depth.msw, dt))
                        decosegment(mix, thisstop.bar+sp, thisstop.bar+sp, dt)
                        ceil_depth = currentCeiling()
                        stoptime += dt
                    scheduleText.append('           Stop at %.1fm for %i min, mix %s' % \
                        (thisstop.msw, math.ceil(stoptime), mix))
                    # current depth, total time of stop, runtime at which we leave, mix
                    schedule.append( (thisstop, math.ceil(stoptime), temp.runtime, mix) )
                
                # ascend to next stop depth
                scheduleText.append('[%.1f min] Ascend from %.1fm to %.1fm' % \
                    (temp.runtime, thisstop.msw, nextstop.msw))
                logger.info("[%.1f min] Ascending to %.2fm, ceiling %.2fm" % \
                    (temp.runtime, nextstop.msw, ceil_depth.msw))
                dt = self.decoDepthInterval.bar / self.ascentSpeed.bar
                decosegment(mix, thisstop.bar+sp, nextstop.bar+sp, dt)
                
            # we have surfaced
            logger.info("[%.1f min] Surfaced, ceiling %.2fm" % (temp.runtime, ceil_depth.msw))

            # calculate decotime and difference with previous iteration
            prev_decotime = decotime
            decotime = temp.runtime - self.runtime
            decodiff = prev_decotime - decotime
            logger.info("runtime %.1f, decotime %.1f (prev %.1f, diff %.1f)" % (
                temp.runtime, decotime, prev_decotime, decodiff))

            # we're done if the Critical Volume Algorithm is off
            if not self.useCriticalVolumeAlgorithm:
                converged = True
                break
            #>nocva
                
            #<cva
            # if the difference is less than the stop time steps, 
            # we concider the profile converged
            if decodiff < self.alignStopTimes:
                converged = True
                logger.info('Profile converged after %i iterations, runtime %.1f min' % \
                    (iteration, temp.runtime))
                break
        
            # update iteration counter for next iteration
            iteration += 1
            if iteration > 100:
                raise Exception("Deco profile doesn't seem to converge. Bug??")
            
            # calculate phase volume times (array)
            spvt = surfacePhaseVolumeTime(self.surfaceMix, self.surfacePressure.bar,
                temp.tissuesHe, temp.tissuesN2, self.halftimes[0], self.halftimes[1])
            phaseVolumeTime = decotime + spvt - time_to_deco_start
            
            # calculate new allowed supersaturation gradients
            prevAllowHe = allowHeFirstStop
            prevAllowN2 = allowN2FirstStop
            logger.info("iter #%i: previous allowed He gradients %s" % (iteration-1, 
                prevAllowHe))
            logger.info("iter #%i: previous allowed N2 gradients %s" % (iteration-1, 
                prevAllowN2))
            if self.useAllowableGradient==1:
                allowHeFirstStop = newAllowableGradient(initialAllowHe, phaseVolumeTime, 
                    self.maxPCrushHe)
                allowN2FirstStop = newAllowableGradient(initialAllowN2, phaseVolumeTime, 
                    self.maxPCrushN2)
            elif self.useAllowableGradient==2:
                # FIXME some tissues??
                allowHeFirstStop = newAllowableGradient2(initialAllowHe, decotime, 
                    phaseVolumeTime, self.maxPCrushHe, 
                    temp.tissuesN2+temp.tissuesHe, self.lastMaxDepth.bar)
                allowN2FirstStop = newAllowableGradient2(initialAllowN2, decotime, 
                    phaseVolumeTime, self.maxPCrushN2, 
                    temp.tissuesN2+temp.tissuesHe, self.lastMaxDepth.bar)
            else:
                raise "Invalid self.useAllowableGradient = %i" % self.useAllowableGradient
            #>cva
            logger.info("iter #%i: new allowed He gradients %s" % (iteration, 
                allowHeFirstStop))
            logger.info("iter #%i: new allowed N2 gradients %s" % (iteration, 
                allowN2FirstStop))
            logger.info("iter #%i: N2 convergence %s" % (iteration, 
                allowN2FirstStop-prevAllowN2))
            logger.info("iter #%i: He convergence %s" % (iteration, 
                allowHeFirstStop-prevAllowHe))
            
        # update our official data for this schedule
        # TODO: record deco profile as a DiveProfile
        #print 'max', temp.maxActualGradients
        #print 'meer he', temp.maxActualGradients - initialAllowHe
        #print 'meer n2', temp.maxActualGradients - initialAllowN2
        self.tissuesHe = temp.tissuesHe
        self.tissuesN2 = temp.tissuesN2 
        self.runtime = temp.runtime
        self.lastIterationCount = iteration
        if firststop:
            self.lastFirstStop = firststop
        else:
            self.lastFirstStop = 0 * msw
        current.log('vpm.iterations', 0, iteration)
        self.lastScheduleText = scheduleText
        self.lastSchedule = schedule
    #>decompress
#>vpm



