#  OpenDeco - An implementation of the Varying Permeability Model used for
#             calculating diving decompression tables
#  Copyright (C) 2007 Konrad Wojas <konrad@opendeco.org>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

#<vpmphysics
import logging
logger = logging.getLogger('opendeco.vpmphysics')

from numpy import exp, sqrt, log, max, maximum, array

from opendeco.units import msw, fsw, atm, mmHg

# used for exponential halftime term
ln2 = log(2)

#>vpmphysics
class ConstVPM_B(object):
    """
    Container for VPM-B constants
    Keeps things tidy and referencing it as const.foo in the code
    helpes keeping it readable.
    Don't reference the class attributes directly, use the const instance.
    """

    def c(value, doc=''):
        """
        Shortcut to define constant properties with docstrings
        """
        f = lambda self: value
        d = "%s %s" % (value, doc)
        return property(f, doc=d)

    gamma = c(17.9,
        "[dyne/cm = 1e-3 N/m] surface tension of bubbles")
    
    gamma_c = c(257, 
        "[dyne/cm = 1e-3 N/m] surface skin compression")

    initialCriticalRadiusN2 = c(0.55, 
        "[micron = 1e-6 m] Critical radius parameter for N2 (0.2-1.35)")
    
    initialCriticalRadiusHe = c(0.45,
        "[micron = 1e-6 m] Critical radius parameter for He (0.2-1.35)")

    regenerationHalftime = c(20160, 
        "min; Bubble regeneration time constant, will only affect long dives")

    lambda_ = c(6500 * fsw, 
        """[fsw*min] This constant absorbs the S, Vcrit, N0, Nsafe, Nactual and
        alpha VPM decompressions constants.""")

    gradientOnsetImperm = c(8.2 * atm,
        "; Pressure gradient over bubble skin at which the skin becomes impermeable")

    pressureOtherGasses = c(102 * mmHg, "; Pressure in bubble of other gasses "
        "like water and CO2 (constant for all ambient pressures)")
    
    waterVaporPressure = c(0.493 * msw,
        "Pressure of water vapor in alveoli (lungs), based on a respiratory quotient of 0.8")


class ConstVPM_A(ConstVPM_B):
    initialCriticalRadiusN2 = 0.80
    initialCriticalRadiusHe = 0.70
    lambda_ = 7500

#<vpmphysics
# make an instance
const = ConstVPM_B()


def conservatismRadiusFactor(conservatism):
    """
    From the HHS Software Corp. V-Planner FAQ:
      Inside the program, the conservatism setting increases the Critical
      Radii of N2/He in the VPM algorithms. From these base values, by this 
      increase: 1 = 5%, 2 = 12%, 3 = 22%, 4 = 35%, 5 = 50%. 
      The Critical Volume setting is on by default.
      See: http://www.hhssoftware.com/v-planner-cde/faqhelp.html#conserv
    conservatism: [int 0-5] conservatism number
    """
    #        +0    +1    +2    +3    +4    +5
    factor = [1, 1.05, 1.12, 1.22, 1.35, 1.50]
    return factor[conservatism]

#<crushingPressure
def crushingPressure(p0, p1, tension0, tension1):
    """
    Calculates the crushing pressure for each compartment for a given descent.
    The calculation is trivial for the permeable regime, but less trivial for
    the impermeable regime.
    p0:       starting pressure of descent
    p1:       ending pressure of descent
    tension0: [array] sum of all gas pressures in compartment at start
    tension1: [array] sum of all gas pressures in compartment at end
    returns:  (cp_he, cp_n2) which are arrays of crushing pressures
    It's up to the calling routine to keep track of the maximum crushing 
    pressures encountered in a dive for the decompressions calculation.
    """
    # - starting gradient (array)
    g0 = p0 - tension0
    # - ending gradient (array)
    g1 = p1 - tension1 #:testlabel
    
    if max(g1) <= const.gradientOnsetImperm.bar:
        # all tissues in permeable regime
        cp_he = g1 
        cp_n2 = g1
    
    else:
        # at least one in impermeable regime
        raise NotImplementedError(), "Nog niet af"
        cp_he = array([0.0] * size)
        cp_n2 = array([0.0] * size)
        size = len(tension0)
        for i in range(size):
            if g1 <= const.gradientOnsetImperm.bar:
                # tissues in permeable regime
                cp_he[i] = g1 
                cp_n2[i] = g1
            else:
                # tissues in impermeable regime
                if g0==const.gradientOnsetImperm.bar:
                    amb_onset = p0
                    g_onset = g0
                elif g0 < const.gradientOnsetImperm.bar:
                    pass
        # FIXME
    return (cp_he, cp_n2)
#>crushingPressure


def radiusAfterGradient(r, dp):
    """
    Calculate the radii after a given gradient,
    r:  [array] initial radii (micron)
    dp: (bar) applied max gradient
    Eq 1 in Y&H paper.
    """
    dp_pa = dp * 1e5 # convert bar to Pa
    micron = 1e-6 # conversion factor for micron to meter
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    logger.debug('r %s, dp %s' % (r, dp))
    r_inv = (1 / (r*micron)) + dp_pa / (2 * (const.gamma_c - const.gamma)*gamma_to_SI)
    new_r =  1 / (r_inv*micron)
    logger.debug('result %s' % new_r)
    return new_r


def initialAllowableGradient(radii):
    """
    Calculate initial allowable gradient for given array of radii
    using Eq 3 in the Y&H paper.
    radii:  [array] (micron) critical radii
    returns [array] (bar) allowed gradients
    If the Critical Volume algorithm isn't used this will provide
    the final (overconservative) decompression schedule.
    """
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    pascal_to_bar = 1e-5
    gamma = const.gamma * gamma_to_SI
    gamma_c = const.gamma_c * gamma_to_SI
    micron = 1e-6 # conversion factor for micron to meter
    G = 2.0 * gamma * (gamma_c - gamma) / (gamma_c * radii * micron)
    return G * pascal_to_bar


def newAllowableGradient(initial_gradient, t_phasevolume, P_crush): #:newAllowableGradient
    """
    Calculate new allowable decogradient for the next Critical Volume Algorithm
    iteration. After a few iterations the value of this will converge. 
    This means we have found the shortest decoprofile with the actual phasevolume
    below or equal to the critical volume.
    initial_gradient: [array] allowable decogradient in first iteration
    t_phasevolume:    decotime + surfacePhaseVolumeTime()
    P_crush:          [array] of max crushing pressures
    Returs:           [array]
    """
    gamma_to_SI = 1e-3 # conversion factor from dyne/cm to N/m
    pascal_to_bar = 1e-5
    bar_to_pascal = 1e5
    gamma = const.gamma * gamma_to_SI
    gamma_c = const.gamma_c * gamma_to_SI
    lambda_ = const.lambda_.pascal

    #<newAllowableGradient
    b = initial_gradient*bar_to_pascal + lambda_ * gamma / ( gamma_c * t_phasevolume )
    c = lambda_ * gamma**2 * P_crush*bar_to_pascal / ( gamma_c**2 * t_phasevolume )
    G = ( b + sqrt(b**2 - 4*c) ) / 2
    G_bar = G * pascal_to_bar
    #>newAllowableGradient

    # critical radii for debugging
    r = 2 * gamma * (gamma_c/gamma) / ( G - P_crush*bar_to_pascal*(gamma/gamma_c) )
    logger.debug('Critical radii (micron): %s' % (r*1e6))

    return G_bar


def newAllowableGradient2(initial_gradient, t_deco, t_phasevolume, P_crush, tensions, p_max):
    """
    Calculate new allowable decogradient for the next Critical Volume Algorithm
    iteration. After a few iterations the value of this will converge. 
    This means we have found the shortest decoprofile with the actual phasevolume
    below or equal to the critical volume.
    Adapted version from 1999 paper FIXME
    initial_gradient: [array] allowable decogradient in first iteration
    t_deco:           decotime + surfacePhaseVolumeTime()
    t_phasevolume:    decotime + surfacePhaseVolumeTime()
    P_crush:          [array] of max crushing pressures
    tensions          [array] compartment tensions at end of dive
    p_max             [array] max pressure in dive
    Returs:           [array]
    FIXME: looks plain wrong!!!
    """
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    pascal_to_bar = 1e-5
    bar_to_pascal = 1e5
    gamma = const.gamma * gamma_to_SI
    gamma_c = const.gamma_c * gamma_to_SI
    lambda_ = const.lambda_.pascal

    newpart = (tensions - p_max) * bar_to_pascal * t_deco / (2 * t_phasevolume)

    b = initial_gradient*bar_to_pascal + lambda_ * gamma / ( gamma_c * t_phasevolume ) \
        - newpart
    c = lambda_ * gamma**2 * P_crush*bar_to_pascal / ( gamma_c**2 * t_phasevolume ) \
        - initial_gradient * newpart
    G = ( b + sqrt(b**2 - 4*c) ) / 2
    G_bar = G * pascal_to_bar

    # critical radii for debugging
    r = 2 * gamma * (gamma_c/gamma) / ( G - P_crush*bar_to_pascal*(gamma/gamma_c) )
    logger.debug('Critical radii (micron): %s' % (r*1e6))

    return G_bar


def newAllowableGradientOriginalYountHoffman(initial_gradient, t_deco, ht, P_crush):
    """
    The old Yount & Hoffman algorithm, unusable for trimix
    FIXME: doc
    Returs: [array]
    """
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    pascal_to_bar = 1e-5
    bar_to_pascal = 1e5
    gamma = const.gamma * gamma_to_SI
    gamma_c = const.gamma_c * gamma_to_SI
    lambda_ = const.lambda_.pascal

    k = ht / ln2 # FIXME klopt dit?
    b = initial_gradient*bar_to_pascal + lambda_ * gamma / ( gamma_c * (t_deco + k) )
    c = lambda_ * gamma**2 * P_crush*bar_to_pascal / ( gamma_c**2 * (t_deco + k) )
    G = ( b + sqrt(b**2 - 4*c) ) / 2
    G_bar = G * pascal_to_bar

    # critical radii for debugging
    r = 2 * gamma * (gamma_c/gamma) / ( G - P_crush*bar_to_pascal*(gamma/gamma_c) )
    logger.debug('Critical radii (micron): %s' % (r*1e6))

    return G_bar


def surfacePhaseVolumeTime(surfaceMix, surfacePressure, tissuesHe, tissuesN2, #:surfacePhaseVolumeTime
                           ht_he, ht_n2):
    """
    Calculates the surface phase volume time according to the algorithm
    used by Erik Baker in his Fortran implementation.
    The extra complexity compared to the original paper by Yount & Hoffman
    is due to the usage of mixed gasses.
    Surface mixes containing Helium currently aren't supported.
    All pressures must be given in bar.
    FIXME: doc
    """
    if surfaceMix.he > 1e-3:
        raise Exception("Surface mix must contain no helium")
    
    surfaceN2 = surfaceMix.n2 * (surfacePressure - const.waterVaporPressure.bar)
    size = len(tissuesHe)
    pvt = array([0.0] * size)
    for i in range(size):
        #<trimix_surface
        n2 = tissuesN2[i]
        he = tissuesHe[i]
        t_n2 = ln2 / ht_n2[i]
        t_he = ln2 / ht_he[i]

        if n2 > surfaceN2:
            pvt[i] = ( he/t_he + (n2 - surfaceN2)/t_n2 ) / (he + n2 - surfaceN2)

        elif (he + n2) >= surfaceN2:
            decay = 1.0/(t_n2 - t_he) * log((surfaceN2 - n2) / he)

            time_to_zero = he/t_he * (1.0 - exp(-t_he * decay)) \
                         + (n2 - surfaceN2)/t_n2 * (1.0 - exp(-t_n2 * decay))

            pvt[i] = time_to_zero / (he + n2 - surfaceN2)
        #>trimix_surface
    return pvt


def ceilings(tissuesHe, tissuesN2, allowHe, allowN2, otherGasses): #:ceilings
    """
    Calculate the ceilings for ascent using the current tissue pressures
    and allowed gradients.

    tissues*:    [array] tissue loadings
    allow*:      [array] allowed supersaturation per tissue
    otherGasses: [float] (bar) pressure of other gasses in bubbles
    returns      [array] of *absolute* ambient ceiling pressures, calling routine will 
                 probably pick the max() one

    TODO: it would be nice to add a prediction of offgassing during ascent,
    the ceiling might be higher than calculated here once the diver gets there.
    Now the calling routine needs to check if it has to stay until a rounded time.
    On the other hand, prediction will not work properly with gas mixes, 
    exotic deco preferences, etc. Lets just keep it this way for now.
    """
    gasloading = tissuesHe + tissuesN2

    # We need to do this element by element, because a gasloading of zero
    # would cause a division by zero. This could happen after breathing oxygen for
    # an extended time.
    size = len(gasloading)
    # tolerated ambient pressure
    tolerated = array([0.0] * size)
    for i in range(size):
        if gasloading[i]>0:
            # normal case, calculate a weighted value
            weighted = ( allowHe[i] * tissuesHe[i] + allowN2[i] * tissuesN2[i] ) \
                       / gasloading[i]
        else:
            # exceptional case; would cause a division by zero, pick the safest one
            # can only happen after breathing oxygen for an extended period
            weighted = min(allowHe[i], allowN2[i])
            # how often does this actually happen?
            logger.info('Found an exceptional case where gas loading is zero in #%i' % i)
        
        tolerated[i] = gasloading[i] + otherGasses - weighted
    
    # If the tolerated ambient pressure is lower than 0, vacuum, then set it to 0
    tolerated = maximum(tolerated, 0.0)
    return tolerated


def boylesLawCompensation(firstStop, nextStop, surfacePressure, allowedFirstStop): #:boylesLawCompensation
    """
    Key feature of VPM-B.
    Boyle's Law compensation, will decrese allowable gradients
    during ascent.
    firstStop, nextStop : DepthUnit
    surfacePressure     : float (bar)
    allowedFirstStop    : [array] allowed supersaturation gradients at first stop
    """
    bar_to_pascal = 1e5
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    gamma = const.gamma * gamma_to_SI

    #<boyle
    # no waterVaporPressure here, not about inspired pressure
    amb_first = (firstStop.bar + surfacePressure) * bar_to_pascal
    amb_next = (nextStop.bar + surfacePressure) * bar_to_pascal
    
    size = len(allowedFirstStop)
    firstradius = array([0.0] * size)
    newradius = array([0.0] * size)
    newgradient = array([0.0] * size)
    firstgradient = allowedFirstStop * bar_to_pascal
    firstradius = (2.0 * gamma) / firstgradient

    for i in range(size):
        r0 = firstradius[i]
        firstradius[i] = r0
        a = amb_next
        b = -2.0 * gamma
        c = amb_first * r0**3 + 2.0 * gamma * r0**2
        low = r0
        high = r0 * (amb_first / amb_next)**(1.0/3.0)
        newradius[i] = radiusRootFinder(a, b, c, low, high)
    
    newgradient = (2.0 * gamma) / newradius
    newgradient_bar = newgradient / bar_to_pascal
    #>boyle

    logger.debug("First gradient (%.1f m): %s" % (firstStop.msw, allowedFirstStop))
    logger.debug("Next gradient (%.1f m): %s" % (nextStop.msw, newgradient_bar))
    logger.debug("First radius (%.1f m): %s" % (firstStop.msw, firstradius*1e6))
    logger.debug("Next radius (%.1f m): %s" % (nextStop.msw, newradius*1e6))

    return newgradient_bar


def radiusRegeneration(pCrush, radius, divetime):
    """
    radius: critical radii
    """
    bar_to_pascal = 1e5
    micron = 1e-6 # conversion factor for micron to meter
    gamma_to_SI = 1e-3 # conversion factor for dyne/cm to N/m
    gamma = const.gamma * gamma_to_SI
    gamma_c = const.gamma_c * gamma_to_SI
    pc = pCrush * bar_to_pascal

    # FIXME: check
    #r = 1.0 / (pc / (2.0*(gamma_c - gamma)) + 1.0 / rc)
    r_cur = radiusAfterGradient(radius, pCrush)
    r_new = radius + (r_cur - radius) * exp(-divetime / const.regenerationHalftime)

    logger.debug("Radius given as parameter: %s" % (radius))
    logger.debug("Radius after applying Pcrush: %s" % (r_cur))
    logger.debug("Radius after regeneration: %s" % (r_new))
    logger.debug("Radius change due to regeneration: %s" % (r_new - r_cur))

    # In order to preserve reference back to the initial critical radii after
    # regeneration, an "adjusted crushing pressure" for the nuclei in each
    # compartment must be computed.  In other words, this is the value of
    # crushing pressure that would have reduced the original nucleus to the
    # to the present radius had regeneration not taken place.  The ratio
    # for adjusting crushing pressure is obtained from algebraic manipulation
    # of the standard VPM equations.  The adjusted crushing pressure, in lieu
    # of the original crushing pressure, is then applied in the VPM Critical
    # Volume Algorithm and the VPM Repetitive Algorithm. 

    adjust_ratio = r_cur * (radius - r_new) / (r_new * (radius - r_cur))
    new_pcrush = pCrush * adjust_ratio
    logger.debug("Pcrush before regeneration: %s" % (pCrush))
    logger.debug("Pcrush after regeneration: %s" % (new_pcrush))
    return new_pcrush


def radiusRootFinder(a, b, c, lowBound, highBound):
    """
    From Erik Bakers Fortran code:
    Purpose: This subroutine is a "fail-safe" routine that combines the
    Bisection Method and the Newton-Raphson Method to find the desired root.
    This hybrid algorithm takes a bisection step whenever Newton-Raphson would
    take the solution out of bounds, or whenever Newton-Raphson is not
    converging fast enough.  Source:  "Numerical Recipes in Fortran 77",
    Cambridge University Press, 1992.
    """
    def function(r):
        """Solve for radius in the form Ar^3 - Br^2 - C = 0"""
        return a * (r**3) - b * (r**2) - c

    def df_dr(r):
        """First derivative"""
        return 3.0 * a * (r**2) - 2.0 * b * r

    f_low = function(lowBound)
    f_high = function(highBound)

    # check for special conditions
    if f_low > 0 and f_high > 0:
        raise Exception("Function for both bounds > 0")
    if f_low < 0 and f_high < 0:
        raise Exception("Function for both bounds < 0")
    
    if f_low==0.0:
        return lowBound
    if f_high==0.0:
        return highBound
    
    if f_low < 0.0:
        r_low = lowBound
        r_high = highBound
    else:
        r_low = highBound
        r_high = lowBound

    ending_radius = 0.5 * (lowBound + highBound)
    last_diff_change = abs(highBound - lowBound)
    diff_change = last_diff_change

    # At this point, the Newton-Raphson Method is applied which uses a function
    # and its first derivative to rapidly converge upon a solution.
    # Note: the program allows for up to 100 iterations.  Normally an exit will
    # be made from the loop well before that number.  If, for some reason, the
    # program exceeds 100 iterations, an exception is raised.
    # When a solution with the desired accuracy is found, exit is made from the
    # loop by returning to the calling program.  The last value of ending
    # radius has been assigned as the solution.    

    f = function(ending_radius)
    df = df_dr(ending_radius)

    for iteration in range(100):
        if (ending_radius-r_high) * (df-f) * (ending_radius-r_low) * (df-f) > 0.0 \
                or abs(2.0 * f) > abs(last_diff_change * df):
            
            last_diff_change = diff_change
            diff_change = 0.5 * (r_high - r_low)
            
            ending_radius = r_low + diff_change
            if r_low==ending_radius:
                return ending_radius
        
        else:
            last_diff_change = diff_change
            diff_change = f/df
            last_ending_radius = ending_radius

            ending_radius = ending_radius - diff_change
            if last_ending_radius==ending_radius:
                return ending_radius
            
        if abs(diff_change) < 1e-12:
            return ending_radius

        f = function(ending_radius)
        df = df_dr(ending_radius)

        if f < 0.0:
            r_low = ending_radius
        else:
            r_high = ending_radius

    # exceeded 100 iterations
    raise Exception("Exceeded 100 iterations")
#>vpmphysics
    

