r"""
Sensitivity analysis of the VPM model.
We will vary several paramers:

 Nitrogen percentage in bottom mixture.
 Helium percentage in bottom mixture.
 Bottom time.
 Maximum depth.
 Descent rate (default 30m/min).
 Ascent rate (default 10m/min).
 Initial critical radius of nitrogen.
 Initial critical radius of helium.
 VPM constant $\lambda$.
 All tissue halftimes.
 Constant pressure of other gasses in nuclei.
 Water vapor pressure in alveoli.
 Surface barometric pressure.
 Water density.
 Pressure of onset of impermeability.
 Surface tension $\gamma$.
 Surface skin compression $\gamma_c$.
"""

from opendeco.units import *

class Const(object):
    gamma = 17.9
    gamma_c = 257 
    initialCriticalRadiusN2 = 0.55 
    initialCriticalRadiusHe = 0.45
    regenerationHalftime = 20160
    lambda_ = 6500 * fsw
    gradientOnsetImperm = 8.2 * atm
    pressureOtherGasses = 102 * mmHg
    waterVaporPressure = 0.493 * msw

# this is fragile
normalconst = Const()
from opendeco import vpmphysics
vpmphysics.const = Const() 

import common

import sys
import os
from numpy import *
import time
import copy

from opendeco import vpm, tracker, mix, tissues
from opendeco.diveseries import *



class SensitivityTester(object):
    
    # default to air
    bottommix = None
    initial_bottommix = mix.Air

    # VPM instance
    vpm = None
    profile = None

    depth = None
    initial_depth = None
    
    time = None
    initial_time = None

    ascentrate = None
    initial_ascentrate = 10
    
    descentrate = None
    initial_descentrate = 30

    kwargs = None
    initial_kwargs = None

    normal = None

    def __init__(self):
        self.initial_kwargs = {}

    def reset(self):
        # reset to clean state
        self.bottommix = self.initial_bottommix
        self.depth = self.initial_depth
        self.time = self.initial_time
        self.kwargs = copy.copy(self.initial_kwargs)
        self.ascentrate = self.initial_ascentrate
        self.descentrate = self.initial_descentrate
        self.profile = None
        for key in dir(normalconst):
            if not key.startswith('_'):
                setattr(vpmphysics.const, key, getattr(normalconst, key))

    def calibrate(self):
        self.reset()
        self.init()
        self.vpm.runProfile(self.profile)
        dt = self.vpm.runtime - self.time
        self.normal = dt
        
    def init(self):
        # initialize, must be called from handlers
        tracker.current.data = {}
        self.vpm = vpm.VPM(**self.kwargs)
        self.profile = common.simpleProfile(self.bottommix, self.depth, 
            self.time, self.descentrate)
        #print self.profile

    def run(self, var, percent):
        assert self.normal, "please calibrate"
        factor = 1.00 + percent/100.0
        print time.time(), var, percent,
        self.reset()
        m = getattr(self, 'vary_'+var)
        m(factor)
        self.vpm.runProfile(self.profile)
        #common.printSchedule(self.vpm)
        dt = self.vpm.runtime - self.time
        change = (dt / self.normal) * 100 - 100
        print '%.2f %.1f' % (dt, change)
        return dt, change

    def vary_mix_n2(self, factor):
        m = self.bottommix
        n2 = m.n2 * factor
        self.bottommix = mix.GasMix(1.00 - n2 - m.he, m.he)
        self.init()

    def vary_mix_he_n2(self, factor):
        # subsitute he for n2
        m = self.bottommix
        he = m.he * factor
        n2 = m.n2 + (m.he - he)
        self.bottommix = mix.GasMix(1.00 - n2 - he, he)
        self.init()

    def vary_mix_he_n2_abs(self, factor):
        # subsitute he for n2 using absolute mix %
        m = self.bottommix
        he = m.he + (factor - 1.0)
        if he < 0:
            he = 0
        n2 = m.n2 + (m.he - he)
        self.bottommix = mix.GasMix(1.00 - n2 - he, he)
        self.init()

    def vary_mix_he(self, factor):
        m = self.bottommix
        he = m.he * factor
        self.bottommix = mix.GasMix(1.00 - m.n2 - he, he)
        self.init()

    def vary_time(self, factor):
        self.time *= factor
        self.init()

    def vary_depth(self, factor):
        self.depth *= factor
        self.init()

    def vary_ascent(self, factor):
        self.vpm.ascentSpeed *= self.ascentrate * factor * msw
        self.init()

    def vary_descent(self, factor):
        self.descentrate *= factor
        self.init()

    def vary_r_n2(self, factor):
        vpmphysics.const.initialCriticalRadiusN2 *= factor
        self.init()

    def vary_r_he(self, factor):
        vpmphysics.const.initialCriticalRadiusHe *= factor
        self.init()

    def vary_lambda(self, factor):
        vpmphysics.const.lambda_ *= factor
        self.init()

    def vary_gamma(self, factor):
        vpmphysics.const.gamma *= factor
        self.init()

    def vary_gamma_c(self, factor):
        vpmphysics.const.gamma_c *= factor
        self.init()

    def vary_othergasses(self, factor):
        vpmphysics.const.pressureOtherGasses *= factor
        self.init()

    def vary_vapor(self, factor):
        vpmphysics.const.waterVaporPressure *= factor
        self.init()

    def vary_surfacepressure(self, factor):
        self.kwargs['surfacePressure'] = 1.013 * factor * bar
        self.init()

    def vary_density(self, factor):
        # TODO
        self.init()


    def vary_onset(self, factor):
        # TODO
        self.init()

    def vary_ht(self, factor):
        self.init()
        he, n2 = tissues.halftimes['buhlmann']
        self.vpm.halftimes = (he * factor, n2* factor)

if __name__=='__main__':
    outbase = common.outbase(__file__)
    st = SensitivityTester()
    st.initial_depth = 60
    st.initial_time = 50
    st.initial_bottommix = mix.Tx(18,45)
    #st.initial_bottommix = mix.Air
    st.initial_kwargs = {
        #'decoMixes' : [(6*msw, mix.Oxygen), (21*msw, mix.EAN(50))],
        'alignStopTimes' : 0.25,
        'decoDepthInterval' : 0.2 * msw
        #'alignStopTimes' : 0.5,
        #'decoDepthInterval' : 0.5 * msw
    }
    st.calibrate()
    
    sym = {
        'mix_n2' : '+--',
        'mix_he' : '+-.',
        'mix_he_n2' : '+:',
        'mix_he_n2_abs' : '+-',
        'time' : 'o-.',
        'depth' : 's-',
        'ascent' : '1:',
        'descent' : '2:',
        'r_n2' : 'x--',
        'r_he' : 'x-.',
        'lambda' : 'D-',
        'gamma' : '<--',
        'gamma_c' : '>--',
        'othergasses' : '4:',
        'vapor' : '3-',
        'surfacepressure' : 'h--',
        'ht' : 'H-',
    }

    tests = [
        'mix_n2',
        'mix_he_n2',
        'mix_he_n2_abs',
    #    'mix_he', 
        'time', 
        'depth',
    #    'ascent',
    #    'descent',
        'r_n2',
    #    'r_he',
        'lambda',
        'gamma',
        'gamma_c',
        'othergasses',
        'vapor',
        'surfacepressure',
    #    'ht',
    ]
    # minder uiteenlopend:
    XXXtests = [
    #    'mix_n2', 
    #    'mix_he', 
    #    'time', 
    #    'depth',
    #    'ascent',
    #    'descent',
        'r_n2',
        'r_he',
        'lambda',
    #    'gamma',
        'gamma_c',
        'othergasses',
        'vapor',
        'surfacepressure',
    #    'ht',
    ]
    tests = [
        'mix_he_n2_abs',
    ]
    #vary_pct = [-20, -10, 10, 20]
    vary_pct = range(-20, 24, 4)
    results = []

    append = ''
    try:
        f = open(sys.argv[1], 'r')
        vary_pct, results = eval(f.read())
        f.close()
        append = '-' + os.path.splitext( os.path.basename(sys.argv[1]) )[0]
    except IndexError:
        for v in tests:
            r = (v, [])
            results.append(r)
            for p in vary_pct:
                if p==0:
                    r[1].append(0)
                else:
                    dt, ch = st.run(v, p)
                    r[1].append(ch)


    print '# depth=%s time=%s gas=%s' % (st.initial_depth, 
        st.initial_time, st.initial_bottommix)
    print '#', time.gmtime()
    print '(', vary_pct, ','
    print results, ')'


    linestyles = ['--', ':', '-.', '-']
    markers = '+o.svx><^Dd1234hHp|_'

    from pylab import *
    i = 0
    j = 0
    for label, res in results:
        i = (i + 1) % len(linestyles)
        j = (j + 1) % len(markers)
        f = markers[j] + linestyles[i]
        f = sym[label]
        plot(vary_pct, res, f+'k')

    gca().set_ylim([-50,50])

    legend([ x[0] for x in results ], loc='best')
    title("VPM sensitivity analysis")
    xlabel('Parameter variation (%)')
    ylabel('Decotime variation (%)')

    common.savefig(outbase+append)
    #common.show()

